/*
 * Copyright (C) 2006 Tommi Saviranta <wnd@iki.fi>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <curses.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <strings.h>
#include <time.h>


#define REFRESH_INTERVAL	30

#define CHR_BORN	'#'
#define	CHR_ALIVE	'#'
#define CHR_TOO_MANY	'o'
#define CHR_TOO_FEW	'x'
#define CHR_NONE	' '

enum {
	COL_BLACK,
	COL_WHITE,
	COL_GREEN,
	COL_BLUE
};

struct coord {
	char chr;
	int age;
	int set;
};

struct change {
	char chr;
	int x;
	int y;
	int set;
};

struct settings {
	int width;
	int height;
	int wrap;
	int aging;
	int old_age;
	int died_age;
	int delay;

	char chr_born;
	char chr_alive;
	char chr_too_many;
	char chr_too_few;
	char chr_none;

	int refresh;
	int alive;
	int size;
};



static struct coord *screen = NULL;
static struct change *change = NULL;



static int
init(int width, int height)
{
	int i;

	screen = malloc((width + 2) * (height + 2) * sizeof(struct coord));
	if (screen == NULL) {
		return -1;
	}

	change = malloc((width * height + 1) * sizeof(change));
	if (change == NULL) {
		free(screen);
		return -1;
	}

	for (i = 0; i < (width + 2) * (height + 2); i++) {
		screen[i].chr = ' ';
		screen[i].age = 0;
		screen[i].set = 0;
	}

	return 0;
}



static void
clean_up(int a)
{
	endwin();
	if (change != NULL) {
		free(change);
		change = NULL;
	}
	if (screen != NULL) {
		free(screen);
		screen = NULL;
	}
	exit(0);
}



static void
wrap_screen(struct settings *settings)
{
	int width = settings->width;
	int height = settings->height;
	int lw = settings->width + 2;
	int i0, i1;
	int x, y;

	i0 = lw;
	for (y = 0; y < height; y++) {
		screen[i0].set = screen[i0 + lw - 2].set;
		screen[i0 + lw - 1].set = screen[i0 + 1].set;
		i0 += lw;
	}

	i0 = 1;
	i1 = (height + 1) * lw + 1;
	for (x = 0; x < width; x++) {
		screen[i0].set = screen[i1 - lw].set;
		screen[i1].set = screen[i0 + lw].set;
		i0++;
		i1++;
	}

	screen[0].set = screen[lw * (height + 1) - 2].set;
	screen[lw - 1].set = screen[lw * height + 1].set;
	screen[lw * (height + 1)].set = screen[lw + lw - 2].set;
	screen[lw * (height + 2) - 1].set = screen[lw + 1].set;
}



static inline void
compute_single(int x, int y, int i, int *set, int lw, struct settings *settings)
{
	int n;
	int is;

	n = screen[i - lw - 1].set +
		screen[i - lw].set +
		screen[i - lw + 1].set +
		screen[i - 1].set +
		screen[i + 1].set +
		screen[i + lw - 1].set +
		screen[i + lw].set +
		screen[i + lw + 1].set;
	is = screen[i].set;

	if (is == 0) {
		if (n == 3) {
			change[*set].x = x;
			change[*set].y = y;
			change[*set].set = 1;
			change[*set].chr = settings->chr_born;
			(*set)++;
		}
	} else {
		if (n > 3) {
			change[*set].x = x;
			change[*set].y = y;
			change[*set].set = 0;
			change[*set].chr = settings->chr_too_many;
			(*set)++;
		} else if (n < 2) {
			change[*set].x = x;
			change[*set].y = y;
			change[*set].set = 0;
			change[*set].chr = settings->chr_too_few;
			(*set)++;
		} else if (settings->aging == 1 &&
				screen[i].age == settings->old_age &&
				rand() % 5 == 0) {
			change[*set].x = x;
			change[*set].y = y;
			change[*set].set = 0;
			change[*set].chr = settings->chr_too_many;
			(*set)++;
		}
	}
}



static void
compute(struct settings *settings)
{
	int width = settings->width;
	int height = settings->height;
	int x, y;
	int lw = width + 2;
	int set = 0;

	if (settings->wrap == 1) {
		wrap_screen(settings);
	}

	for (y = 0; y < height; y++) {
		int i = (y + 1) * (width + 2) + 1;
		for (x = 0; x < width; x++) {
			compute_single(x, y, i, &set, lw, settings);
			i++;
		}
	}

	change[set].set = -1;
}



static inline void
print_colored(int x, int y, int is, int ir, int lw, int refresh,
		struct settings *settings)
{
	if (screen[is].set == 0) {
		if (screen[is].age < settings->died_age) {
			if (refresh == 0) {
				screen[is].age++;
			}
			move(y, x);
			attrset(COLOR_PAIR(COL_BLUE) | A_DIM);
			addch(screen[is].chr);
		} else if (screen[is].age == settings->died_age) {
			screen[is].chr = settings->chr_none;
			move(y, x);
			attrset(COLOR_PAIR(COL_BLACK));
			addch(screen[is].chr);
			screen[is].age++;
		}
	} else {
		if (screen[is].age == 0) {
			if (refresh == 0) {
				screen[is].age++;
			}
			move(y, x);
			attrset(COLOR_PAIR(COL_GREEN) | A_BOLD);
			addch(screen[is].chr);
		} else if (screen[is].age == 1 || refresh == 1) {
			if (refresh == 0) {
				screen[is].age++;
			}
			screen[is].chr = settings->chr_alive;
			move(y, x);
			attrset(COLOR_PAIR(COL_GREEN) | A_DIM);
			addch(screen[is].chr);
		} else if (settings->old_age != 0) {
			if (refresh == 0) {
				screen[is].age++;
			}
		}
	}
}



static inline void
age_single(int x, int y, int is, int ir, int lw, struct settings *settings)
{
	print_colored(x, y, is, ir, lw, 0, settings);
}



static void
age(struct settings *settings)
{
	int width = settings->width;
	int height = settings->height;

	int x, y;
	int lw = width + 2;
	int ir;

	ir = 0;
	for (y = 0; y < height; y++) {
		int is = y * lw + lw + 1;
		for (x = 0; x < width; x++) {
			age_single(x, y, is, ir, lw, settings);
			is++;
			ir++;
		}
	}
}



static void
commit(struct settings *settings)
{
	int width = settings->width;
	int aging = settings->aging;

	int s = 0;
	int lw = width + 2;

	while (change[s].set != -1) {
		int i;
		if (settings->aging == 0) {
			mvaddch(change[s].y, change[s].x, change[s].chr);
		}

		i = change[s].y * lw + lw + change[s].x + 1;
		settings->alive += (change[s].set - screen[i].set);

		screen[i].set = change[s].set;
		screen[i].age = 0;
		screen[i].chr = change[s].chr;
		s++;
	}

	if (aging == 1) {
		age(settings);
	}

	refresh();
	usleep(settings->delay * 1000);
}



static inline void
splot(int x, int y, int lw, int s, char c)
{
	int i = y * lw + x;
	screen[i].set = s;
	screen[i].age = 0;
	screen[i].chr = c;
}



static void
create_garbage(struct settings *settings)
{
	int n = rand() % settings->size / 100;
	int lw = settings->width + 2;

	while (n > 0) {
		int x = rand() % settings->width;
		int y = rand() % settings->height;

		splot(x, y, lw, 1, settings->chr_born);
		n--;
	}
}



static void
create_static_0(struct settings *settings)
{
	int n = rand() % 10;
	int lw = settings->width + 2;

	while (n > 0) {
		int x = rand() % (settings->width - 2);
		int y = rand() % (settings->height - 2);

		splot(x, y, lw, 1, settings->chr_born);
		splot(x + 1, y, lw, 1, settings->chr_born);
		splot(x, y + 1, lw, 1, settings->chr_born);
		splot(x + 1, y + 1, lw, 1, settings->chr_born);
		n--;
	}
}



static void
create_static_1(struct settings *settings)
{
	int n = rand() % 4;
	int lw = settings->width + 2;

	while (n > 0) {
		int x = rand() % (settings->width - 4);
		int y = rand() % (settings->height - 4);

		splot(x + 1, y, lw, 1, settings->chr_born);
		splot(x + 2, y, lw, 1, settings->chr_born);
		splot(x, y + 1, lw, 1, settings->chr_born);
		splot(x + 3, y + 1, lw, 1, settings->chr_born);
		splot(x, y + 2, lw, 1, settings->chr_born);
		splot(x + 3, y + 2, lw, 1, settings->chr_born);
		splot(x + 1, y + 3, lw, 1, settings->chr_born);
		splot(x + 2, y + 3, lw, 1, settings->chr_born);
		n--;
	}
}



static void
create_traffic_lights(struct settings *settings)
{
	int n = rand() % 4;
	int lw = settings->width + 2;

	while (n > 0) {
		int x = rand() % (settings->width - 3);
		int y = rand() % (settings->height - 3);

		splot(x, y + 1, lw, 1, settings->chr_born);
		splot(x + 1, y + 1, lw, 1, settings->chr_born);
		splot(x + 2, y + 1, lw, 1, settings->chr_born);
		n--;
	}
}



static void
create_glider(struct settings *settings)
{
	int lw = settings->width + 2;
	int x = rand() % (settings->width - 3);
	int y = rand() % (settings->height - 3);

	int t = rand() % 4;

	switch (t) {
	case 0:
		splot(x, y, lw, 1, settings->chr_born);
		splot(x + 1, y, lw, 1, settings->chr_born);
		splot(x + 2, y, lw, 1, settings->chr_born);
		splot(x + 2, y + 1, lw, 1, settings->chr_born);
		splot(x + 1, y + 2, lw, 1, settings->chr_born);
		break;
	case 1:
		splot(x + 2, y, lw, 1, settings->chr_born);
		splot(x, y + 1, lw, 1, settings->chr_born);
		splot(x + 2, y + 1, lw, 1, settings->chr_born);
		splot(x + 1, y + 2, lw, 1, settings->chr_born);
		splot(x + 2, y + 2, lw, 1, settings->chr_born);
		break;
	case 2:
		splot(x + 1, y, lw, 1, settings->chr_born);
		splot(x, y + 1, lw, 1, settings->chr_born);
		splot(x, y + 2, lw, 1, settings->chr_born);
		splot(x + 1, y + 2, lw, 1, settings->chr_born);
		splot(x + 2, y + 2, lw, 1, settings->chr_born);
		break;
	case 3:
		splot(x, y, lw, 1, settings->chr_born);
		splot(x + 1, y, lw, 1, settings->chr_born);
		splot(x, y + 1, lw, 1, settings->chr_born);
		splot(x + 2, y + 1, lw, 1, settings->chr_born);
		splot(x, y + 2, lw, 1, settings->chr_born);
		break;
	}
}



static void
plot_random(struct settings *settings, int force)
{
	int limit;

	limit = (int) log((double)
			(1 + (settings->alive << 5) / settings->size));

	if (force == 1 ||
			((rand() % (int) log(settings->size) > limit) &&
				(rand() % 30 == 0))) {
		int t = rand() % 5;

		switch (t) {
		case 0:
			create_garbage(settings);
			break;
		case 1:
			create_static_0(settings);
			break;
		case 2:
			create_static_1(settings);
			break;
		case 3:
			create_traffic_lights(settings);
			break;
		case 4:
			create_glider(settings);
			break;
		}
	}
}



static void
total_redraw(struct settings *settings)
{
	int x, y;
	int lw = settings->width + 2;

	clear();
	if (settings->aging == 1) {
		int ir = 0;
		for (y = 0; y < settings->height; y++) {
			int is = y * lw + lw + 1;
			for (x = 0; x < settings->width; x++) {
				print_colored(x, y, is, ir, lw, 1, settings);
				is++;
				ir++;
			}
		}
	} else {
		for (y = 0; y < settings->height; y++) {
			int is = y * lw + lw + 1;
			for (x = 0; x < settings->width; x++) {
				mvaddch(y, x, screen[is].chr);
				is++;
			}
		}
	}

	refresh();
}



int
main(int argc, char **argv)
{
	struct settings settings;
	struct sigaction sv;
	int t;
	time_t next;

	settings.alive = 0;

	settings.refresh = 30;
	settings.delay = 50;
	settings.wrap = 1;
	settings.aging = 1;
	settings.died_age = 10;
	if (settings.aging == 1) {
		settings.old_age = 200;
		settings.chr_born = CHR_BORN;
		settings.chr_alive = CHR_ALIVE;
		settings.chr_too_many = CHR_TOO_MANY;
		settings.chr_too_few = CHR_TOO_FEW;
	} else {
		settings.chr_born = settings.chr_alive = CHR_ALIVE;
		settings.chr_too_many = settings.chr_too_few = CHR_NONE;
	}
	settings.chr_none = CHR_NONE;
	
	sigemptyset(&sv.sa_mask);
	sv.sa_flags = 0;
	sv.sa_handler = clean_up;

	srand(time(NULL));

	sigaction(SIGINT, &sv, NULL);

	initscr();
	keypad(stdscr, TRUE);
	nonl();
	cbreak();
	noecho();
	curs_set(0);

	getmaxyx(stdscr, settings.height, settings.width);
	t = init(settings.width, settings.height);

	settings.size = settings.width * settings.height;

	start_color();
	init_pair(COL_BLACK, COLOR_BLACK, COLOR_BLACK);
	init_pair(COL_WHITE, COLOR_WHITE, COLOR_BLACK);
	init_pair(COL_GREEN, COLOR_GREEN, COLOR_BLACK);
	init_pair(COL_BLUE, COLOR_BLUE, COLOR_BLACK);

	change[0].set = -1;

	/* traffic light */
	/*
	create_traffic_lights(&settings);
	create_traffic_lights(&settings);
	create_traffic_lights(&settings);
	create_traffic_lights(&settings);
	create_traffic_lights(&settings);
	*/

	plot_random(&settings, 1);
	plot_random(&settings, 1);
	plot_random(&settings, 1);
	total_redraw(&settings);

	if (settings.refresh > 0) {
		next = time(NULL) + settings.refresh;
	} else {
		next = 0; /* keep some ccs happy */
	}
	while (t == 0) {
		if (settings.refresh > 0) {
			time_t now;
			now = time(NULL);
			if (now >= next) {
				total_redraw(&settings);
				next = time(NULL) + settings.refresh;
			}
		}
		compute(&settings);
		commit(&settings);
		plot_random(&settings, 0);
	}

	clean_up(0);

	return 0;
}
